/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/



/** \file RobinBC.h */

#ifndef AMDIS_ROBINBC_H
#define AMDIS_ROBINBC_H

#include "BoundaryCondition.h"
#include "AbstractFunction.h"
#include "DOFMatrix.h"

namespace AMDiS {

  /** 
   * \ingroup Assembler
   *
   * \brief
   * Sub class of BoundaryCondition. Implements Robin and Neumann boundary conditions.
   * The flux in normal direction is given by \f$ A\nabla u\cdot \nu\phi =: j = j_0 - \alpha u \f$ where
   * \f$ j_0 \f$ and \f$ alpha \f$ are functions evaluated at world coordinates
   * and \f$ u \f$ is the problem solution.
   */
  class RobinBC : public BoundaryCondition
  {
  public:
    /// Constructor. \f$ j \f$ and \f$ alpha \f$ are given as AbstractFunction objects.
    RobinBC(BoundaryType type,
	    AbstractFunction<double, WorldVector<double> > *j,
	    AbstractFunction<double, WorldVector<double> > *alpha,
	    const FiniteElemSpace *rowFeSpace,
	    const FiniteElemSpace *colFeSpace = NULL);

    /// Constructor. \f$ j \f$ and \f$ alpha \f$ are given as DOFVectors.
    RobinBC(BoundaryType type,
	    DOFVectorBase<double> *j,
	    DOFVectorBase<double> *alpha,
	    const FiniteElemSpace *rowFeSpace,
	    const FiniteElemSpace *colFeSpace = NULL);

    /// Constructor. \f$ j \f$ and \f$ alpha \f$ are given as Operator objects.
    RobinBC(BoundaryType type,
	    Operator* jOp, Operator* alphaOp,
	    const FiniteElemSpace *rowFeSpace,
	    const FiniteElemSpace *colFeSpace = NULL);

    /// Implements BoundaryCondition::fillBoundaryCondition();
    virtual void fillBoundaryCondition(DOFMatrix* matrix,
				       ElInfo* elInfo,
				       const DegreeOfFreedom* dofIndices,
				       const BoundaryType* localBound,
				       int nBasFcts);
  
    /// Implements BoundaryCondition::fillBoundaryCondition();
    virtual void fillBoundaryCondition(DOFVectorBase<double>* vector, 
				       ElInfo* elInfo,
				       const DegreeOfFreedom* dofIndices,
				       const BoundaryType* localBound,
				       int nBasFcts);

    /// Implements BoundaryCondition::boundResidual();
    virtual double boundResidual(ElInfo *elInfo, 
				 DOFMatrix *matrix,
				 const DOFVectorBase<double> *dv);

  protected:
    /// Surface operators for each element side for the Neumann part.
    DimVec<SurfaceOperator*>* neumannOperators;

    /// Surface operators for each element side for the Robin part.
    DimVec<SurfaceOperator*>* robinOperators;

    VectorOfFixVecs<DimVec<double> >**coords;
  };


  class NeumannBC : public RobinBC
  {
  public:
    NeumannBC(BoundaryType type,
	      AbstractFunction<double, WorldVector<double> > *j,
	      const FiniteElemSpace *rowFeSpace,
	      const FiniteElemSpace *colFeSpace = NULL)
      : RobinBC(type, j, NULL, rowFeSpace, colFeSpace)
    {}

    NeumannBC(BoundaryType type,
	      DOFVectorBase<double> *j,
	      const FiniteElemSpace *rowFeSpace,
	      const FiniteElemSpace *colFeSpace = NULL)
      : RobinBC(type, j, NULL, rowFeSpace, colFeSpace)
    {}
  };

}

#endif
